import asyncio
import contextlib
import logging
from starlette.middleware import Middleware
import importlib
import json
import os
from pathlib import Path
from types import ModuleType

from pydantic import BaseModel
from relax.injection import (
    Injected,
    injectable,
    COMPONENTS_CACHE_FILE,
    _COMPONENT_NAMES,
)
from relax.config import BaseConfig
from starlette.websockets import WebSocket, WebSocketDisconnect


import inspect
from dataclasses import Field, _MISSING_TYPE
from enum import StrEnum, auto
from functools import partial, wraps
from inspect import Parameter, signature
from typing import (
    Annotated,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    ClassVar,
    Concatenate,
    Generic,
    Literal,
    Mapping,
    Protocol,
    Self,
    Sequence,
    Type,
    TypedDict,
    TypeVar,
    get_args,
    get_origin,
)
from starlette.datastructures import URL, UploadFile

import starlette.requests
import starlette.types
from starlette.applications import Starlette
from starlette.authentication import requires
import starlette.responses
from starlette.routing import Route
from typing_extensions import ParamSpec

import relax.html

QueryStr = Annotated[str, "query_param"]
QueryInt = Annotated[int, "query_param"]
PathInt = Annotated[int, "path_param"]
PathStr = Annotated[str, "path_param"]

P = ParamSpec("P")
T = TypeVar("T")

CLIENTS: set[WebSocket] = set()
IMPORTS: dict[str, ModuleType] = {}


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


DataclassT = TypeVar("DataclassT", bound=DataclassInstance)

Method = Literal["GET", "POST", "PUT", "DELETE", "PATCH"]

logger = logging.getLogger(__name__)


class HTMLResponse(starlette.responses.HTMLResponse):
    def __init__(
        self,
        content: relax.html.Element,
        status_code: int = 200,
        headers: Mapping[str, str] | None = None,
    ) -> None:
        super().__init__(content.render(), status_code, headers)


class AuthScope(StrEnum):
    Authenticated = auto()


class Scope(TypedDict):
    from_htmx: bool


class Request(starlette.requests.Request, Generic[T]):
    user: T  # type: ignore
    scope: Scope  # type: ignore

    def __init__(self, base_request: starlette.requests.Request) -> None:
        super().__init__(base_request.scope, base_request._receive, base_request._send)

    def url_of(
        self,
        func: Callable[Concatenate["Request", P], Awaitable[Any]],
        *_: P.args,
        **kwargs: P.kwargs,
    ) -> URL:
        return self.url_for(func.__name__, **kwargs)

    def url_wrapper(
        self,
        func: Callable[Concatenate["Request", P], Awaitable[Any]],
    ) -> Callable[P, URL]:
        def inner(*_: P.args, **kwargs: P.kwargs) -> URL:
            return self.url_for(func.__name__, **kwargs)

        return inner

    @contextlib.asynccontextmanager
    async def parse_form(
        self,
        data_shape: Type[DataclassT],
    ) -> AsyncGenerator[DataclassT, None]:
        fields: dict[str, Any] = {}
        async with self.form() as form:
            for name, field in data_shape.__dataclass_fields__.items():
                try:
                    if field.type is UploadFile:
                        fields[name] = form[name]
                    else:
                        fields[name] = field.type(form[name])
                except KeyError:
                    if isinstance(field.default, _MISSING_TYPE):
                        raise
                    fields[name] = field.default
            yield data_shape(**fields)


class UserType(Protocol):
    username: str
    display_name: str
    is_authenticated: bool


def get_annotated(param: Parameter) -> Any:
    annotation = param.annotation
    if get_origin(annotation) is Annotated:
        return get_args(annotation)

    if get_origin(getattr(get_origin(annotation), "__value__", None)) is Annotated:
        return (
            get_args(annotation)[0],
            get_args(getattr(get_origin(annotation), "__value__", None))[1],
        )

    try:
        if get_origin(get_args(annotation)[0]) is Annotated:
            return get_args(get_args(annotation)[0])
    except IndexError:
        return None

    return None


def load_views() -> dict | None:
    with COMPONENTS_CACHE_FILE.open("r") as f:
        try:
            raw_data = json.load(f)
        except TypeError:
            logger.exception("failed parsing JSON data")
            raise

    updated_views = {}

    logger.warning("loaded data")
    try:
        for id, element in raw_data.items():
            fn_path = element["path"]
            fn_values = element["data"]
            fn_module, fn_name = fn_path.rsplit(".", 1)
            if fn_module not in IMPORTS:
                IMPORTS[fn_module] = importlib.import_module(fn_module)
            fn = getattr(IMPORTS[fn_module], fn_name)

            # TODO: if we can't find the function referenced in the data, drop it
            for name, param in signature(fn).parameters.items():
                try:
                    if (
                        param.default is not Injected
                        and isinstance(fn_values.get(name), dict)
                        and issubclass(param.annotation, BaseModel)
                    ):
                        model_obj = fn_values[name]
                        fn_values[name] = param.annotation(**(model_obj))
                except TypeError:
                    pass

            updated_views[id] = fn(**fn_values).render()
    except Exception as e:  # noqa: BLE001
        logger.warning("failed loading views: %s", repr(e))
        return None
    return updated_views


async def websocket_endpoint(websocket: WebSocket) -> None:
    await websocket.accept()
    logger.warning("got new websocket connection")
    CLIENTS.add(websocket)
    try:
        while True:
            data = await websocket.receive_text()
            logger.warning("got new data: %s", data)
    except WebSocketDisconnect:
        CLIENTS.remove(websocket)


class ViewContext:
    def __init__(self) -> None:
        self._app: Starlette | None = None
        self.endpoints: dict[Callable, Any] = {}
        self.path_functions: dict[Callable, Callable] = {}

    def attach_to_app(self, app: Starlette) -> Self:
        self._app = app
        return self

    def add_endpoint(self, sig: Any, endpoint: Callable) -> None:
        self.endpoints[sig] = endpoint

    def endpoint(self, sig: Type[T]) -> T:
        return self.endpoints[sig]

    def add_path_function(self, func: Callable) -> None:
        if self._app is None:
            msg = "App instance not set on ViewContext"
            raise ValueError(msg)
        self.path_functions[func] = partial(self._app.url_path_for, func.__name__)

    def url_of(
        self,
        func: Callable[Concatenate["Request", P], Awaitable[Any]],
    ) -> Callable[P, URL]:
        return self.path_functions[func]  # type: ignore


class App(Starlette):
    def __init__(
        self,
        view_context: ViewContext,
        config: BaseConfig,
        debug: bool = False,
        middleware: Sequence[Middleware] | None = None,
        lifespan: starlette.types.Lifespan["App"] | None = None,
    ) -> None:
        super().__init__(debug=debug, middleware=middleware, lifespan=lifespan)
        self.view_context = view_context.attach_to_app(self)
        self.config = config

    def add_router(self, router: "Router") -> None:
        for route in router.routes:
            self.routes.append(route)
            self.view_context.add_path_function(route.endpoint)

    def listen_to_template_changes(self) -> None:
        print("Listening to template changes for hot-module replacement")
        self.listen_task = asyncio.create_task(self._listen_to_template_changes())

    async def _listen_to_template_changes(self) -> None:
        with contextlib.suppress(FileNotFoundError):
            self.config.RELOAD_SOCKET_PATH.unlink()
        print("gonna listen on socket")
        self.reload_server = await asyncio.start_unix_server(
            intermediary_hot_replace_templates,
            self.config.RELOAD_SOCKET_PATH,
        )
        print("started listening on socket: ", self.reload_server.is_serving())


class BaseRouter(Protocol):
    ...


class Router(BaseRouter):
    def __init__(self) -> None:
        self.routes: list[Route] = []

    def path_function(  # noqa: ANN201
        self,
        method: Method,
        endpoint: str,
        auth_scopes: list[AuthScope] | None = None,
    ):
        if auth_scopes is None:
            auth_scopes = []

        def decorator(func):  # noqa: ANN001, ANN202
            @wraps(func)
            @requires(auth_scopes)
            async def inner(request: starlette.requests.Request):  # noqa: ANN202
                request = Request(request)
                params: dict[str, Any] = {}
                request.scope["from_htmx"] = (
                    request.headers.get("HX-Request", False) == "true"
                )
                for param_name, param in signature(func).parameters.items():
                    args = get_annotated(param)
                    if args and args[1] == "query_param":
                        if (
                            param_value := request.query_params.get(param_name)
                        ) is None:
                            if param.default is inspect._empty:
                                msg = (
                                    f"parameter {param_name} from function "
                                    f"{func.__name__} has no default value "
                                    "and was not provided in the request"
                                )
                                raise TypeError(msg)
                            params[param_name] = param.default
                        else:
                            # TODO: also allow something like
                            # \ Annotated[Path | None, "query_param"] = Path("/")
                            params[param_name] = args[0](param_value)
                    elif args and args[1] == "path_param":
                        if (param_value := request.path_params.get(param_name)) is None:
                            if param.default is inspect._empty:
                                msg = (
                                    f"parameter {param_name} from function "
                                    f"{func.__name__} has no default value "
                                    "and was not provided in the request"
                                )
                                raise TypeError(msg)
                            params[param_name] = param.default
                        else:
                            params[param_name] = args[0](param_value)
                return await injectable(func)(request, **params)

            # TODO: maybe make the name file + fn_name?
            # TODO: also, error out when finding a duplicate name
            self.routes.append(
                Route(endpoint, inner, methods=[method], name=func.__name__),
            )

            return inner

        return decorator


def update_js_constants(config: BaseConfig) -> None:
    with config.JS_CONSTANTS_PATH.open("w") as f:
        f.write("export const CONSTANTS = {\n")
        for name in _COMPONENT_NAMES:
            f.write(f'   {name.upper().replace("-", "_")}_CLASS: "{name}",\n')
        f.write("}")


async def intermediary_hot_replace_templates(
    sr: asyncio.StreamReader,
    _: asyncio.StreamWriter,
) -> None:
    try:
        data = await sr.read(1024)
        result = json.loads(data)
        if result["event_type"] == "update_views":
            await hot_replace_templates(result["data"])
    except Exception as e:  # noqa: BLE001
        print(e)  # noqa: T201


async def hot_replace_templates(changed_paths: list[str]) -> None:
    try:
        for str_path in changed_paths:
            file_path = Path(str_path)
            str_path = str_path.removesuffix(file_path.suffix)
            str_path = str_path.replace(os.sep, ".")
            if str_path in IMPORTS:
                IMPORTS[str_path] = importlib.reload(IMPORTS[str_path])
            else:
                IMPORTS[str_path] = importlib.import_module(str_path)
                IMPORTS[str_path] = importlib.reload(IMPORTS[str_path])

        logger.warning("reloaded changes")
        new_views = load_views()
        logger.warning("loaded views")
        if new_views is not None:
            for client in CLIENTS:
                await client.send_text(
                    json.dumps({"event_type": "update_views", "data": new_views}),
                )
            logger.warning("updated browser")
        else:
            logger.warning("no data to update server with")
    except Exception as e:  # noqa: BLE001
        print(e)  # noqa: T201



import json
from pathlib import Path
import warnings
from collections.abc import Callable
from functools import wraps
from inspect import _ParameterKind, signature
from typing import Any, Awaitable, Hashable, ParamSpec, TypeVar, Protocol, Self

from relax.html import Component, Element

"""
A simple dependency injection framework.

Decorate functions that need to be injected with `@injectable` and
annotate the parameters that need to be injected with `Injected`.

Note: Injected parameters must be keyword-only.

Example:
```python
@injectable
def do_print(
    foo: str = "default",
    *,
    bar: Annotated[str, Injected],
) -> None:
    print(foo)
    print(bar)
```
Running `do_print()` print
```
default
weird
```
"""


class MissingDependencyError(Exception):
    ...


class IncorrectInjectableSignatureError(Exception):
    ...


class DoubleInjectionError(Exception):
    ...


_INJECTS: dict[Hashable, object] = {}


class _Injected:
    ...


Injected: Any = _Injected


_P = ParamSpec("_P")
_T = TypeVar("_T")

COMPONENTS_CACHE_FILE = Path("/tmp/relax_components.json")
_COMPONENT_NAMES: list[str] = []


class Jsonable(Protocol):
    def model_dump_json(self) -> str:
        ...

    @classmethod
    def model_validate_json(
        cls,
        json_data: str | bytes | bytearray,
    ) -> Self:
        ...


def inject_into_kwargs(func: Callable, kwargs: Any) -> None:
    for name, sig in signature(func).parameters.items():
        if sig.default is _Injected:
            if sig.kind is not _ParameterKind.KEYWORD_ONLY:
                msg = (
                    f"Injected parameter {name} in "
                    f"{func.__name__} must be keyword-only"
                )
                raise IncorrectInjectableSignatureError(msg)
            if kwargs.get(name) is not None:
                pass
            elif sig.annotation in _INJECTS:
                kwargs.update({name: _INJECTS[sig.annotation]})
            else:
                msg = (
                    "Missing dependency for "
                    f"{name}: {sig.annotation} in {func.__name__}"
                )
                raise MissingDependencyError(msg)


def injectable(func: Callable[_P, Awaitable[_T]]) -> Callable[_P, Awaitable[_T]]:
    @wraps(func)
    async def inner(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        inject_into_kwargs(func, kwargs)
        return await func(*args, **kwargs)

    return inner


def injectable_sync(func: Callable[_P, _T]) -> Callable[_P, _T]:
    @wraps(func)
    def inner(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        inject_into_kwargs(func, kwargs)
        return func(*args, **kwargs)

    return inner


def add_injectable(annotation: Hashable, injectable: object) -> None:
    if annotation in _INJECTS:
        msg = f"Injectable {annotation} already added"
        raise DoubleInjectionError(msg)
    _INJECTS[annotation] = injectable


def retrieve_injectable(annotation: Hashable) -> object:
    return _INJECTS[annotation]

def clear_injections() -> None:
    return _INJECTS.clear()



def component(
    key: Callable[..., str] | str | None = None,
) -> Callable[[Callable[_P, Element]], Callable[_P, Component]]:
    def decorator(
        func: Callable[_P, Element],
    ) -> Callable[_P, Component]:
        component_name = func.__name__.replace("_", "-")
        # TODO: don't do this in dev, or find a way to make it useful
        if component_name in _COMPONENT_NAMES:
            msg = f"Component {component_name} already registered"
            warnings.warn(msg, stacklevel=1)
        _COMPONENT_NAMES.append(component_name)

        @wraps(func)
        def inner(**kwargs: Jsonable) -> Component:
            if isinstance(key, str):
                key_val = key
                elem_id = f"{component_name}-{key_val}"
            elif key:
                lsig = signature(key)
                lambda_args = []
                for p_name in lsig.parameters:
                    lambda_args.append(kwargs[p_name])
                key_val = key(*lambda_args)
                elem_id = f"{component_name}-{key_val}"
            else:
                elem_id = component_name
            new_func = injectable_sync(func)
            if "id" in signature(func).parameters:
                # TODO: don't set the id if it was provided in the kwargs already
                func_call_result = new_func(id=elem_id, **kwargs)
            else:
                func_call_result = new_func(**kwargs)
            data = {key: to_json(val) for (key, val) in kwargs.items()}
            # TODO: find a way to do this only in dev mode
            try:
                with COMPONENTS_CACHE_FILE.open() as f:
                    current_views = json.load(f)
            except (TypeError, FileNotFoundError):
                current_views = {}
            view_key = f"{func.__module__}.{func.__name__}"
            current_views[elem_id] = {
                "path": view_key,
                "data": data,
                "signature": str(signature(func)),
            }
            with COMPONENTS_CACHE_FILE.open("w") as f:
                json.dump(current_views, f)
            return func_call_result.set_id(elem_id).classes([component_name])

        return inner

    return decorator


def to_json(obj: Any) -> str:
    if hasattr(obj, "model_dump_json"):
        obj = obj.model_dump(mode="json")
    return obj



import contextlib
import time
from socket import AF_UNIX, SOCK_STREAM, socket
import uvicorn.config
from uvicorn.supervisors.watchfilesreload import WatchFilesReload
import uvicorn
import json
from pathlib import Path

from relax.config import BaseConfig
from relax.injection import COMPONENTS_CACHE_FILE


from typing import (
    Callable,
)

MAX_CONNECT_RETRIES = 5


class RelaxReload(WatchFilesReload):
    base_config: BaseConfig

    def __init__(
        self,
        config: uvicorn.config.Config,
        target: Callable[[list[socket] | None], None],
        sockets: list[socket],
        base_config: BaseConfig,
    ) -> None:
        super().__init__(config, target, sockets)
        self.base_config = base_config
        self._reload_socket: socket | None = None

    def should_restart(self) -> list[Path] | None:
        try:
            changed_paths = super().should_restart()
            if changed_paths is None:
                return None
            changed_templates: list[Path] = []
            changed_app_files: list[Path] = []
            for change in changed_paths:
                try:
                    Path(change).relative_to(self.base_config.TEMPLATES_DIR)
                    changed_templates.append(Path(change).relative_to(Path.cwd()))
                # we raise ValueError if the path is not in the templates dir
                except ValueError:
                    changed_app_files.append(change)
            if len(changed_app_files) > 0:
                print("changed app files: ", changed_app_files)
                return changed_app_files
            if len(changed_templates) > 0:
                print("changed templates: ", changed_templates)
                self._update_templates(changed_templates)
        except Exception as e:
            print("failed to check for restart: %s" % e)
            raise
        return None

    def shutdown(self) -> None:
        if self.reload_socket is not None:
            self.reload_socket.close()
        return super().shutdown()

    def run(self, *, remove_components_cache_on_startup: bool = True) -> None:
        if remove_components_cache_on_startup:
            with contextlib.suppress(FileNotFoundError):
                COMPONENTS_CACHE_FILE.unlink()
            with COMPONENTS_CACHE_FILE.open("w") as f:
                json.dump({}, f)
        return super().run()

    def get_new_reload_socket(self, retries: int = 0) -> socket:
        self.reload_socket = socket(AF_UNIX, SOCK_STREAM)
        self.reload_socket.settimeout(5)
        try:
            self.reload_socket.connect(str(self.base_config.RELOAD_SOCKET_PATH))
        except ConnectionRefusedError:
            if retries >= MAX_CONNECT_RETRIES:
                raise
            time.sleep(1)
            return self.get_new_reload_socket(retries + 1)
        return self.reload_socket

    def _update_templates(self, changed_templates: list[Path]) -> None:
        try:
            if self._reload_socket is None:
                self.get_new_reload_socket()
        except ConnectionRefusedError:
            print("can't connect to reload socket, not updating templates")
            return
        changes = json.dumps(
            {
                "event_type": "update_views",
                "data": [str(change) for change in changed_templates],
            },
        ).encode()
        self.reload_socket.send(changes)
        self.reload_socket.close()


def start_app(
    app_path: str,
    config: BaseConfig,
    *,
    host: str = "127.0.0.1",
    port: int | None = None,
    reload: bool = False,
    log_level: str = "info",
) -> None:
    if port is None:
        port = config.PORT

    server_config = uvicorn.Config(
        app=app_path,
        host=host,
        port=port,
        log_level=log_level,
        factory=True,
    )

    server = uvicorn.Server(server_config)

    if reload:
        reload_config = uvicorn.Config(
            app=app_path,
            host=host,
            port=port,
            factory=True,
            reload=True,
            log_level=log_level,
        )
        sock = reload_config.bind_socket()
        reloader = RelaxReload(
            reload_config,
            target=server.run,
            sockets=[sock],
            base_config=config,
        )
        reloader.run()
    else:
        server.run()



import warnings
import sys
from pathlib import Path
from collections.abc import Iterable
from html import escape
from typing import Literal, Protocol, Self, Sequence, TypeVar

from starlette.datastructures import URL

HREFTarget = Literal["_blank", "_self", "_parent", "_top"]

InputType = Literal[
    "text",
    "email",
    "password",
    "number",
    "date",
    "file",
    "hidden",
    "checkbox",
    "radio",
    "button",
]

ButtonType = Literal["button", "submit", "reset"]

HTMXRequestType = Literal["get", "post", "put", "delete"]


class InvalidHTMLError(Exception):
    ...


T = TypeVar("T")


class Element(Protocol):
    _parent: "Tag | None"
    name: str

    def render(self) -> str:
        ...

    def set_id(self, value: str) -> Self:
        ...

    @property
    def id(self) -> str | None:
        ...

    def classes(self, classes: list[str]) -> Self:
        ...

    def attrs(self, attrs: dict) -> Self:
        ...

    def hx_get(
        self,
        target: str | URL,
        hx_encoding: Literal["multipart/form-data"] | None = None,
        hx_target: str | Self | None = None,
        hx_swap: Literal[
            "innerHTML",
            "outerHTML",
            "beforebegin",
            "afterbegin",
            "beforeend",
            "afterend",
            "delete",
            "none",
        ]
        | None = None,
        **kwargs: str,
    ) -> Self:
        ...


class Component(Element, Protocol):
    @property
    def id(self) -> str:
        ...


class SelfClosingTag(Element):
    name: str

    def __init__(
        self,
        *,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
    ) -> None:
        self._text: str = ""
        self._attributes: dict = {}
        self._classes: list[str] = []
        if classes:
            self.classes(classes)
        if attrs:
            self.attrs(attrs)
        if id:
            self.set_id(id)
        if hyperscript:
            self.hyperscript(hyperscript)
        self._parent: Tag | None = None

    def render(self) -> str:
        return f"<{self.name} {self._render_attributes()} />"

    def _render_attributes(self) -> str:
        attributes = [
            f'{escape(key)}="{escape(str(value).strip())}"'.strip()
            for key, value in self._attributes.items()
        ]
        if self._classes:
            attributes.append(
                f'class="{" ".join([escape(klass) for klass in self._classes])}"',
            )
        if attributes:
            return " " + " ".join(attributes).strip()
        return ""

    def classes(self, classes: list[str]) -> Self:
        self._classes.extend(classes)
        return self

    def hyperscript(self, command: str) -> Self:
        self._attributes["_"] = command
        return self

    def attrs(self, attrs: dict) -> Self:
        self._attributes.update(attrs)
        return self

    def set_id(self, value: str) -> Self:
        self._attributes["id"] = value
        return self

    @property
    def id(self) -> str | None:
        return self._attributes["id"]

    def _htmx(
        self,
        request_type: HTMXRequestType,
        target: str | URL,
        hx_encoding: Literal["multipart/form-data"] | None = None,
        hx_target: str | Self | None = None,
        hx_swap: Literal[
            "innerHTML",
            "outerHTML",
            "beforebegin",
            "afterbegin",
            "beforeend",
            "afterend",
            "delete",
            "none",
        ]
        | None = None,
        **kwargs: str,
    ) -> Self:
        self._attributes["hx-" + request_type] = target
        if hx_encoding:
            self._attributes["hx-encoding"] = hx_encoding
        if isinstance(hx_target, SelfClosingTag):
            try:
                self._attributes["hx-target"] = hx_target._attributes["id"]
            except KeyError as exc:
                msg = (
                    f"Target element {hx_target.__class__.__name__} "
                    "must have an id attribute"
                )
                raise InvalidHTMLError(msg) from exc
        elif isinstance(hx_target, str):
            self._attributes["hx-target"] = hx_target
        if hx_swap:
            self._attributes["hx-swap"] = hx_swap
        for key, value in kwargs.items():
            attr = key.replace("_", "-")
            self._attributes[attr] = value
        return self

    def hx_get(
        self,
        target: str | URL,
        hx_encoding: Literal["multipart/form-data"] | None = None,
        hx_target: str | Self | None = None,
        hx_swap: Literal[
            "innerHTML",
            "outerHTML",
            "beforebegin",
            "afterbegin",
            "beforeend",
            "afterend",
            "delete",
            "none",
        ]
        | None = None,
        **kwargs: str,
    ) -> Self:
        return self._htmx("get", target, hx_encoding, hx_target, hx_swap, **kwargs)

    def hx_post(
        self,
        target: str | URL,
        hx_encoding: Literal["multipart/form-data"] | None = None,
        hx_target: str | Self | None = None,
        hx_swap: Literal[
            "innerHTML",
            "outerHTML",
            "beforebegin",
            "afterbegin",
            "beforeend",
            "afterend",
            "delete",
            "none",
        ]
        | None = None,
        **kwargs: str,
    ) -> Self:
        return self._htmx("post", target, hx_encoding, hx_target, hx_swap, **kwargs)

    def hx_put(
        self,
        target: str | URL,
        hx_encoding: Literal["multipart/form-data"] | None = None,
        hx_target: str | Self | None = None,
        hx_swap: Literal[
            "innerHTML",
            "outerHTML",
            "beforebegin",
            "afterbegin",
            "beforeend",
            "afterend",
            "delete",
            "none",
        ]
        | None = None,
        **kwargs: str,
    ) -> Self:
        return self._htmx("put", target, hx_encoding, hx_target, hx_swap, **kwargs)

    def hx_delete(
        self,
        target: str | URL,
        hx_encoding: Literal["multipart/form-data"] | None = None,
        hx_target: str | Self | None = None,
        hx_swap: Literal[
            "innerHTML",
            "outerHTML",
            "beforebegin",
            "afterbegin",
            "beforeend",
            "afterend",
            "delete",
            "none",
        ]
        | None = None,
        **kwargs: str,
    ) -> Self:
        return self._htmx("delete", target, hx_encoding, hx_target, hx_swap, **kwargs)


class Tag(SelfClosingTag):
    def __init__(
        self,
        *,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
        text: str | None = None,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        self._children: list[Element] = []
        if text is not None:
            self.text(text)

    def render(self) -> str:
        return (
            f"<{self.name}{self._render_attributes()}>"
            f"{self._render_children()}"
            f"</{self.name}>"
        )

    def _render_children(self) -> str:
        return escape(self._text) + "".join(
            [child.render() for child in self._children],
        )

    def text(self, text: str) -> Self:
        self._text = text
        if self._children:
            msg = "Cannot have text and children"
            raise InvalidHTMLError(msg)
        return self

    def insert(
        self,
        *children: Sequence[Element] | Element | None,
        append: bool = True,
    ) -> Self:
        if self._text:
            msg = "Cannot have text and children"
            raise InvalidHTMLError(msg)

        # since we are so lenient with what we accept
        # sieve arguments until we get just a list of elements
        final_list: list[Element] = []
        for child in children:
            if child is None:
                continue
            if isinstance(child, Iterable):
                for sub_child in child:
                    final_list.append(sub_child)
                    sub_child._parent = self
            else:
                final_list.append(child)
                child._parent = self
        if append:
            self._children.extend(final_list)
        else:
            self._children = final_list
        return self


class Fragment(Tag):
    name = "<>"

    def __init__(
        self,
        children: Sequence[Element | None] | None,
    ) -> None:
        super().__init__()
        self.insert(children)

    def render(self) -> str:
        return super()._render_children()


class div(Tag):
    name = "div"


class main(Tag):
    name = "main"


class progress(Tag):
    name = "progress"


class nav(Tag):
    name = "nav"


class p(Tag):
    name = "p"


class span(Tag):
    name = "span"


class body(Tag):
    name = "body"


class button(Tag):
    name = "button"

    def __init__(
        self,
        *,
        type: ButtonType | None = None,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
        text: str | None = None,
    ) -> None:
        super().__init__(
            classes=classes,
            attrs=attrs,
            id=id,
            hyperscript=hyperscript,
            text=text,
        )
        if type:
            self._attributes["type"] = type
        else:
            self._attributes["type"] = "button"


class form(Tag):
    name = "form"

    def __init__(
        self,
        *,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
        action: str | None = None,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        if action is not None:
            self._attributes["action"] = action


class i(Tag):
    name = "i"


class a(Tag):
    name = "a"

    def __init__(
        self,
        href: str | URL,
        target: HREFTarget | None = None,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
        text: str | None = None,
    ) -> None:
        super().__init__(
            classes=classes,
            attrs=attrs,
            id=id,
            hyperscript=hyperscript,
            text=text,
        )
        self._attributes["href"] = href
        if target:
            self._attributes["target"] = target


class li(Tag):
    name = "li"

    def render(self) -> str:
        if not self._parent:
            warnings.warn(
                f'"{self.name}" element should be a child of "ul" or "ol"',
                stacklevel=2,
            )

        if self._parent and self._parent.name not in ["ul", "ol"]:
            warnings.warn(
                f'"{self.name}" element should be a child of "ul" or "ol"',
                stacklevel=2,
            )
        return super().render()


class ul(Tag):
    name = "ul"
    # TODO: warning when children are not "li"


class label(Tag):
    name = "label"

    def __init__(
        self,
        _for: str | None = None,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        if _for is not None:
            self._attributes["for"] = _for


class svg(Tag):
    name = "svg"


class path(Tag):
    name = "path"


class select(Tag):
    name = "select"

    def __init__(
        self,
        name: str,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        self._attributes["name"] = name


class option(Tag):
    name = "option"

    def __init__(
        self,
        value: str,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        self._attributes["value"] = value


class input(SelfClosingTag):
    name = "input"

    def __init__(
        self,
        name: str,
        type: InputType,
        value: str | None = None,
        placeholder: str | None = None,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        self._attributes["name"] = name
        self._attributes["type"] = type
        if value:
            self._attributes["value"] = value
        if placeholder:
            self._attributes["placeholder"] = placeholder
        if disabled is True:
            self._attributes["disabled"] = True


class img(SelfClosingTag):
    name = "img"

    def __init__(
        self,
        src: str,
        alt: str,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        self._attributes["alt"] = alt
        self._attributes["src"] = src


class video(Tag):
    name = "video"

    def __init__(
        self,
        src: str,
        *,
        controls: bool = False,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
    ) -> None:
        super().__init__(id=id, classes=classes, attrs=attrs, hyperscript=hyperscript)
        self._attributes["src"] = src
        if controls:
            self._attributes["controls"] = "true"


class textarea(Tag):
    name = "textarea"

    def __init__(
        self,
        name: str,
        type: InputType,
        placeholder: str | None = None,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        self._attributes["name"] = name
        self._attributes["type"] = type
        if placeholder:
            self._attributes["placeholder"] = placeholder
        if disabled is True:
            self._attributes["disabled"] = True


class meta(SelfClosingTag):
    name = "meta"

    def __init__(
        self,
        *,
        charset: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ) -> None:
        if charset and (name or content):
            msg = "Cannot have charset and name/content"
            raise InvalidHTMLError(msg)
        if charset:
            super().__init__(attrs={"charset": charset})
        else:
            super().__init__(attrs={"name": name, "content": content})


class link(SelfClosingTag):
    name = "link"

    def __init__(self, *, href: str, rel: str, type: str | None = None) -> None:
        super().__init__(attrs={"href": href, "rel": rel})
        if type is not None:
            self._attributes["type"] = type


class title(Tag):
    name = "title"

    def __init__(self, name: str) -> None:
        super().__init__()
        self.text(name)


class style(Tag):
    name = "style"

    def __init__(self, stylesheet: str) -> None:
        super().__init__()
        self.text(stylesheet)


class script(Tag):
    name = "script"

    def __init__(
        self,
        js: str | None = None,
        src: str | None = None,
        attrs: dict | None = None,
        *,
        type: Literal["importmap", "module", "speculationrules"] | None = None,
        defer: bool = False,
    ) -> None:
        super().__init__()
        if js is not None:
            self.text(js)
        elif src is not None:
            self._attributes["src"] = src
        else:
            msg = "<script> element must have js or src"
            raise InvalidHTMLError(msg)
        if attrs is not None:
            self._attributes.update(attrs)
        if defer:
            self._attributes["defer"] = "true"
        if type is not None:
            self._attributes["type"] = type

    # sorry mate can't help you escape that
    # your risk
    def _render_children(self) -> str:
        return self._text + "".join([child.render() for child in self._children])


class head(Tag):
    name = "head"
    # TODO: ensure only meta | link | title | style | script elements can be inserted


class html(Tag):
    name = "html"

    def __init__(
        self,
        lang: str,
        classes: list[str] | None = None,
        attrs: dict | None = None,
        id: str | None = None,
        hyperscript: str | None = None,
    ) -> None:
        super().__init__(classes=classes, attrs=attrs, id=id, hyperscript=hyperscript)
        self._attributes["lang"] = lang

    # TODO: ensure only head | body can be inserted

    def render(self) -> str:
        return "<!DOCTYPE html>" + super().render()


class aside(Tag):
    name = "aside"

class details(Tag):
    name = "details"

class summary(Tag):
    name = "summary"

class dialog(Tag):
    name = "dialog"

class h1(Tag):
    name = "h1"


class h2(Tag):
    name = "h2"


class h3(Tag):
    name = "h3"


class h4(Tag):
    name = "h4"

class h5(Tag):
    name = "h5"

def hmr_script() -> list[script]:
    file_path = sys.modules[__name__].__file__
    if file_path is None:
        msg = "Cannot find current file path"
        raise RuntimeError(msg)
    current_path = Path(file_path)
    with (current_path.parent / "js" / "hmr_reload.js").open() as f:
        return [
            script(
                js=f.read(),
            ),
            script(src="https://unpkg.com/idiomorph@0.3.0"),
        ]



import inspect
from typing import Any, Callable, ParamSpec, TypeVar, get_args
from inspect import signature

P = ParamSpec("P")
T = TypeVar("T")


def check(func: Callable[..., None]):  # noqa: ANN201
    renames = {}
    sig = signature(func)
    for name, val in sig.parameters.items():
        annotation = get_args(val.annotation)
        fixture_name = annotation[1].__name__
        renames[name] = fixture_name
    return _rename_parameters(func, renames)


def _rename_parameters(
    func: Callable[..., None],
    rename_dict: dict,
) -> Callable[..., None]:
    original_sig = inspect.signature(func)
    new_params = [
        param.replace(name=rename_dict.get(param.name, param.name))
        for param in original_sig.parameters.values()
    ]
    new_sig = original_sig.replace(parameters=new_params)

    def wrapper(*args: Any, **kwargs: Any):  # noqa: ANN202
        bound_args = new_sig.bind(*args, **kwargs)
        return func(*bound_args.args, **bound_args.kwargs)

    wrapper.__signature__ = new_sig  # type: ignore[reportFunctionMemberAccess]

    return wrapper



from typing import Literal, TypeAlias

import starlette.config
from pathlib import Path

ENV: TypeAlias = Literal["dev", "prod"]


def cast_to_env(value: str) -> Literal["dev", "prod"]:
    if value == "dev":
        return "dev"
    if value == "prod":
        return "prod"
    msg = f"Invalid value: {value}"
    raise ValueError(msg)


def cast_abs_path(value: str) -> Path:
    return Path(value).absolute()


class BaseConfig:
    def __init__(self) -> None:
        self.config = starlette.config.Config()

        self.TEMPLATES_DIR = self.config("TEMPLATES_DIR", cast=cast_abs_path)
        self.ENV = self.config("ENV", default="prod", cast=cast_to_env)
        self.PORT = self.config("PORT", default=443, cast=int)
        self.RELOAD_SOCKET_PATH = self.config(
            "RELOAD_SOCKET_PATH",
            default=Path("/tmp/relax-reload"),
            cast=cast_abs_path,
        )
        self.JS_CONSTANTS_PATH = self.config(
            "JS_CONSTANTS_PATH",
            default=Path("static/js/constants.js"),
            cast=Path,
        )

    def __repr__(self) -> str:
        rep = self.__dict__.copy()
        del rep["config"]
        return rep.__repr__()
